unit Photo;

// Description:	32bpp fast DIB object and windowed image control
// Author:			Richard Shotbolt

// {$HINTS OFF}

interface

uses
  Windows,
  Messages,
  WinTypes,
  UITypes,
  Types,
  SysUtils,
  Classes,
  Graphics,
  Controls,
  Forms,
  Dialogs,
  Math,
  JPEG,
  GIFImg,
  PNGImage,
  MMSystem;

const
  // Define transparent black
  clTransBlack = $02000000;

type

  // resampling support types
  TRGBInt = record
    R, G, B: Integer;
  end;

  PRGBWord = ^TRGBWord;

  TRGBWord = record
    R, G, B: Word;
  end;

  PRGBAWord = ^TRGBAWord;

  TRGBAWord = record
    R, G, B, A: Word;
  end;

  PBGR = ^TBGR;

  TBGR = packed record
    B, G, R: Byte;
  end;

  PBGRA = ^TBGRA;

  TBGRA = packed record
    B, G, R, A: Byte;
  end;

  PRGB = ^TRGB;

  TRGB = packed record
    R, G, B: Byte;
  end;

  PRGBA = ^TRGBA;

  TRGBA = packed record
    R, G, B, A: Byte;
  end;

  PPixelArray = ^TPixelArray;
  TPixelArray = array [0 .. 0] of TBGR;

  PRGB4 = ^TRGB4;

  TRGB4 = packed record
    case Integer of
      0:
        (Color: Cardinal); // Access as integer
      1:
        (Value: Single); // Access as single
      2:
        ( // Access as RGB
          rgbBlue: Byte;
          rgbGreen: Byte;
          rgbRed: Byte;
          rgbReserved: Byte;
        );
      3:
        ( // Access as BGR
          bgrRed: Byte;
          bgrGreen: Byte;
          bgrBlue: Byte;
          bgrReserved: Byte;
        );
      4:
        ( // Access as HSV
          hsvValue: Byte;
          hsvSaturation: Byte;
          hsvHue: Byte;
          hsvIntensity: Byte;
        );
  end;

  TFloatColor = packed record
    fBlue: Single;
    fGreen: Single;
    fRed: Single;
  end;

  TWordColor = packed record
    wBlue: Word;
    wGreen: Word;
    wRed: Word;
  end;

  PFloatColor = ^TFloatColor;

  TFloatColorArray = array [0 .. 0] of TFloatColor;
  PFloatColorArray = ^TFloatColorArray;

  TLine = array [0 .. 0] of TRGB4;
  PLine = ^TLine;

  TFLine = array [0 .. 0] of TFloatColor;
  PFLine = ^TFLine;

  TCalcs = array [0 .. 0] of Integer;
  PCalcs = ^TCalcs;

  TSingleArray = array [0 .. 0] of Single;
  PSingleArray = ^TSingleArray;

type
  // Color difference record
  TColorDiff = record
    Color: TRGB4;
    Diff: Integer;
  end;

const
  SER_MONO = 0;
  SER_RGB = 100;

type
  TSERHeader = packed record
    FileID: array [0 .. 13] of AnsiChar; // 'LUCAM-RECORDER'
    LuID: Integer; // Lumenera camera series ID
    ColorID: Integer; // Color format code
    LittleEndian: Integer; // 0=BigEndian, 1=LittleEndian
    ImageWidth: Integer;
    ImageHeight: Integer;
    PixelDepth: Integer; // Bit depth; 2 bytes if >8
    FrameCount: Integer;
    Observer: array [0 .. 39] of AnsiChar;
    Instrument: array [0 .. 39] of AnsiChar;
    Telescope: array [0 .. 39] of AnsiChar;
    StartTime: Int64;
    StartTimeUTC: Int64;
  end;

type
  TSERFile = class(TObject)
  private
    FBytesPerFrame: Integer;
    FBytesTotal: Int64;
    FFrames: TMemoryStream;
    FHeader: TSERHeader;
    procedure SetFileID(Value: AnsiString);
    function GetFileID: AnsiString;
    procedure SetLuID(Value: Integer);
    function GetLuID: Integer;
    procedure SetColorID(Value: Integer);
    function GetColorID: Integer;
    procedure SetLittleEndian(Value: Integer);
    function GetLittleEndian: Integer;
    procedure SetImageWidth(Value: Integer);
    function GetImageWidth: Integer;
    procedure SetImageHeight(Value: Integer);
    function GetImageHeight: Integer;
    procedure SetPixelDepth(Value: Integer);
    function GetPixelDepth: Integer;
    procedure SetFrameCount(Value: Integer);
    function GetFrameCount: Integer;
    procedure SetObserver(Value: AnsiString);
    function GetObserver: AnsiString;
    procedure SetInstrument(Value: AnsiString);
    function GetInstrument: AnsiString;
    procedure SetTelescope(Value: AnsiString);
    function GetTelescope: AnsiString;
    procedure SetStartTime(Value: Int64);
    function GetStartTime: Int64;
    procedure SetStartTimeUTC(Value: Int64);
    function GetStartTimeUTC: Int64;
    function GetByteCount: Integer;
  protected
  public
    constructor Create;
    destructor Destroy; override;
    function AddFrame: Pointer;
    procedure DeleteFrames;
    function GetFrame(FrameNo: Integer): Pointer;
    procedure SaveToFile(const FN: string);
    procedure LoadFromFile(const FN: string);
    property BytesPerFrame: Integer read FBytesPerFrame;
    property BytesTotal: Int64 read FBytesTotal;
    property Header: TSERHeader read FHeader write FHeader;
    property FileID: AnsiString read GetFileID write SetFileID;
    property LuID: Integer read GetLuID write SetLuID;
    property ColorID: Integer read GetColorID write SetColorID;
    property LittleEndian: Integer read GetLittleEndian write SetLittleEndian;
    property ImageWidth: Integer read GetImageWidth write SetImageWidth;
    property ImageHeight: Integer read GetImageHeight write SetImageHeight;
    property PixelDepth: Integer read GetPixelDepth write SetPixelDepth;
    property FrameCount: Integer read GetFrameCount write SetFrameCount;
    property Observer: AnsiString read GetObserver write SetObserver;
    property Instrument: AnsiString read GetInstrument write SetInstrument;
    property Telescope: AnsiString read GetTelescope write SetTelescope;
    property StartTime: Int64 read GetStartTime write SetStartTime;
    property StartTimeUTC: Int64 read GetStartTimeUTC write SetStartTimeUTC;
  end;

type
  TPhoto = class;

  TPhotoBitmap = class(TObject)
  private
    { Private declarations }
    FBits: Pointer;
    FBMPInfo: PBitmapInfo;
    FBMPChanged: Boolean;
    FColor: TColor;
    FCalcs: PCalcs;
    FCanvas: TCanvas;
    FDC: HDC;
    FFCalcs: PCalcs;
    FHandle: HBITMAP;
    FHeight: Integer;
    FWidth: Integer;
    FFloatColors: PFloatColor;
    FNoFloatColors: Boolean;
    FMonochrome: Boolean;
    FPixelCount: Integer;
    FPixelDepth: Integer;
    FRowInc: Integer;
    FFRowInc: Integer;
    FSize: Integer;
    FFSize: Integer;
    FParent: TPhoto;
    FSERFile: TSERFile;
    FTransparentColor: TColor;
    FVisible: Boolean;
    OldBitmap: HBITMAP;
    SL: TStringList;
    procedure SetColor(Value: TColor);
    procedure SetHeight(Value: Integer);
    procedure SetWidth(Value: Integer);
    procedure SetPixel(X, Y: Integer; Clr: TColor); inline;
    function GetPixel(X, Y: Integer): TColor; inline;
    procedure SetPixelF(X, Y: Integer; Clr: TFloatColor);
    function GetPixelF(X, Y: Integer): TFloatColor;
    procedure SetLine(Y: Integer; Line: Pointer);
    function GetLine(Y: Integer): Pointer;
    procedure SetLineF(Y: Integer; Line: Pointer);
    function GetLineF(Y: Integer): Pointer;
    procedure SetMonochrome(Value: Boolean);
    procedure SetPixelDepth(Value: Integer);
  protected
    { Protected declarations }
  public
    { Public declarations }
    constructor Create;
    destructor Destroy; override;
    procedure NewDIBSection;
    procedure AddSERFrame;
    procedure Assign(Value: TPhotoBitmap);
    function AverageFloatColors: TFloatColor;
    procedure Clear;
    procedure ClearF;
    procedure ClearSERFrames;
    procedure ColorDifference(Dst: TPhotoBitmap);
    procedure ConvolveF(Dims: Integer; Kernel: array of Single; Scale: Single);
    procedure CopyRectPixels(Dst: TPhotoBitmap;
      DstX, DstY, SrcX, SrcY, W, H: Integer);
    function CreateCopy: TPhotoBitmap;
    procedure FlipX;
    procedure FlipY;
    procedure FastLine(x1, y1, x2, y2: Integer);
    procedure DespeckleF;
    procedure Fill(Value: TColor);
    procedure FillTransparent;
    function GetSERFrame(FrameNo: Integer; Dest: TPhotoBitmap): Boolean;
    procedure MakeCompatible(Dest: TPhotoBitmap);
    procedure RadialCorrect(Amount: Single);
    procedure RadialCorrect2(K: Single);
    procedure Rotate(OX, OY: Integer; Angle: Single);
    procedure SaveSERFile(FN: string);
    procedure SetSize(WD, HT: Integer);
    property ScanLines[Y: Integer]: Pointer read GetLine write SetLine;
    property ScanLinesF[Y: Integer]: Pointer read GetLineF write SetLineF;
    procedure FillFloatColors(Source: TPhotoBitmap; DstScale, SrcScale: Single);
    procedure FillFloatFactors(Source: TPhotoBitmap; Value: TFloatColor;
      Mult: Single = 1);
    procedure ApplyFloatFactors(Source: TPhotoBitmap);
    procedure ShowFloatColors(Dest: TPhotoBitmap; Mult: Single = 1);
    property Bits: Pointer read FBits write FBits;
    property Calcs: PCalcs read FCalcs write FCalcs;
    property Canvas: TCanvas read FCanvas write FCanvas;
    property Color: TColor read FColor write SetColor;
    property FloatCalcs: PCalcs read FFCalcs write FFCalcs;
    property FloatColors: PFloatColor read FFloatColors write FFloatColors;
    property Handle: HBITMAP read FHandle;
    property Height: Integer read FHeight write SetHeight;
    property Monochrome: Boolean read FMonochrome write SetMonochrome;
    property NoFloatColors: Boolean read FNoFloatColors write FNoFloatColors;
    property Parent: TPhoto read FParent write FParent;
    property PixelCount: Integer read FPixelCount;
    property PixelDepth: Integer read FPixelDepth write SetPixelDepth;
    property Pixels[X, Y: Integer]: TColor read GetPixel write SetPixel;
    property PixelsF[X, Y: Integer]: TFloatColor read GetPixelF write SetPixelF;
    property Size: Integer read FSize;
    property SizeF: Integer read FFSize;
    property TransparentColor: TColor read FTransparentColor
      write FTransparentColor;
    property Visible: Boolean read FVisible write FVisible default True;
    property Width: Integer read FWidth write SetWidth;
  end;

  TPhotoBitmapArray = array of TPhotoBitmap;
  PPhotoBitmapArray = ^TPhotoBitmapArray;

  TPhoto = class(TCustomControl)
  private
    FAllowDragImage: Boolean;
    FAllowResizeImage: Boolean;
    FBitmap: TPhotoBitmap;
    FColor: TColor;
    FDeferPaint: Boolean;
    FLayerCount: Integer;
    FLayers: TPhotoBitmapArray;
    FNoFloatColors: Boolean;
    FTransColor: TColor;
    FOnBeforePaint: TNotifyEvent;
    FOnAfterPaint: TNotifyEvent;
    NoErase: Boolean;
    procedure SetLayerCount(Value: Integer);
    procedure SetLayer(Idx: Integer; BMP: TPhotoBitmap);
    function GetLayer(Idx: Integer): TPhotoBitmap;
    procedure SetBitmap(Value: TPhotoBitmap);
    procedure SetColor(Value: TColor);
    procedure WMGetDlgCode(var Msg: TWMGetDlgCode); message WM_GETDLGCODE;
  protected
    // procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y:
    // Integer); override;
    procedure WndProc(var Message: TMessage); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure Hide;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
      X, Y: Integer); override;
    function DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
      MousePos: TPoint): Boolean; override;
    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint)
      : Boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint)
      : Boolean; override;
    procedure Paint; override;
    procedure Show;
    property AllowDragImage: Boolean read FAllowDragImage write FAllowDragImage;
    property AllowResizeImage: Boolean read FAllowResizeImage
      write FAllowResizeImage;
    property TransparentColor: TColor read FTransColor write FTransColor
      default clTransBlack;
    property LayerCount: Integer read FLayerCount write SetLayerCount default 1;
    property Layers[Idx: Integer]: TPhotoBitmap read GetLayer
      write SetLayer; default;
    property NoFloatColors: Boolean read FNoFloatColors write FNoFloatColors;
    property Bitmap: TPhotoBitmap read FBitmap write SetBitmap;
    // property Active: Integer read GetActive write SetActive;
  published
    property Align;
    property Canvas;
    property Color: TColor read FColor write SetColor;
    property DragCursor;
    property DragMode;
    property Enabled;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnAfterPaint: TNotifyEvent read FOnAfterPaint write FOnAfterPaint;
    property OnBeforePaint: TNotifyEvent read FOnBeforePaint
      write FOnBeforePaint;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnKeyDown;
    property OnKeyUp;
    property OnKeyPress;
    property OnMouseWheelDown;
    property OnMouseWheelUp;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
  end;

function RGB2HSV(Color: TRGB4): TRGB4;
procedure RGB3To4(Dst, Src: Pointer; Size: Integer);
function RGBGR(Value: TColor): TRGB4; overload;
function RGBGR(Value: TRGB4): TColor; overload;
function ColorDiff(var Col1, Col2: TRGB4): Integer;
function MinOf(Value1, Value2: Integer): Integer; overload;
function MinOf(Value1, Value2: Single): Single; overload;
function MaxOf(Value1, Value2: Integer): Integer; overload;
function MaxOf(Value1, Value2: Single): Single; overload;
function ClampByte(Value: Integer): Byte; overload;
function ClampByte(Value: Single): Byte; overload;
function ClampWord(Value: Single; MaxValue: Word): Word;
function WithinBounds(X, Y: Integer; Rect: TRect): Boolean; overload;
function WithinBounds(Point: TPoint; Rect: TRect): Boolean; overload;

// Register as a component
procedure Register;

implementation

// =============================================================================

function ClampByte(Value: Integer): Byte;
// ensures Value is in the range 0..255, values < 0 are clamped to 0 and values > 255 are clamped to 255
asm
  OR EAX, EAX
  JNS @@positive
  XOR EAX, EAX
  RET
@@positive:
  CMP EAX, 255
  JBE @@OK
  MOV EAX, 255
@@OK:
end;

function ClampByte(Value: Single): Byte;
// ensures Value is in the range 0..255, values < 0 are clamped to 0 and values > 255 are clamped to 255
begin
  Result := Round(Value);
  if (Value < 0) then
    Result := 0
  else if (Value > 255) then
    Result := 255;
end;

// =============================================================================

function ClampWord(Value: Single; MaxValue: Word): Word;
begin
  Result := Round(Value);
  if (Value < 0) then
    Result := 0
  else if (Value > MaxValue) then
    Result := MaxValue;
end;

// =============================================================================

procedure IntsToSingles(Src: PInteger; Dst: PSingle; Size: Integer);
// Convert memory block to array of singles
var
  I: Integer;
begin
  try
    for I := 0 to Size - 1 do
    begin
      Dst^ := Src^;
      Inc(Src);
      Inc(Dst);
    end;
  except
  end;
end;

// =============================================================================

procedure SinglesToInts(Src: PSingle; Dst: PInteger; Size: Integer);
// Convert array of singles to memory block of integers
var
  I: Integer;
begin
  try
    for I := 0 to Size - 1 do
    begin
      Dst^ := ClampByte(Src^);
      Inc(Src);
      Inc(Dst);
    end;
  except
  end;
end;

// =============================================================================

procedure RGB3To4(Dst, Src: Pointer; Size: Integer);
// Write an RGB4 block with RGBTriples
// Very efficient - check out the CPU window!
var
  I: Integer;
  SP: PRGBTriple;
  DP: PRGB4;
begin
  try
    SP := Src;
    DP := Dst;
    for I := 0 to (Size div 4) - 1 do
    begin
      DP^.Color := Cardinal(Pointer(SP)^) and $FFFFFF;
      Inc(SP);
      Inc(DP);
    end;
  except
  end;
end;

// =============================================================================

function MinOf(Value1, Value2: Integer): Integer;
// Return lesser of two values
begin
  if Value1 <= Value2 then
    Result := Value1
  else
    Result := Value2;
end;

// =============================================================================

function MinOf(Value1, Value2: Single): Single;
// Return lesser of two values
begin
  if Value1 <= Value2 then
    Result := Value1
  else
    Result := Value2;
end;

// =============================================================================

function MaxOf(Value1, Value2: Integer): Integer;
// Return greater of two values
begin
  if Value1 >= Value2 then
    Result := Value1
  else
    Result := Value2;
end;

// =============================================================================

function MaxOf(Value1, Value2: Single): Single;
// Return greater of two values
begin
  if Value1 >= Value2 then
    Result := Value1
  else
    Result := Value2;
end;

// =============================================================================

// Returns True if the point X, Y lies within Rect

function WithinBounds(X, Y: Integer; Rect: TRect): Boolean;
begin
  Result := (X >= Rect.Left) and (X <= Rect.Right) and (Y >= Rect.Top) and
    (Y <= Rect.Bottom);
end;

function WithinBounds(Point: TPoint; Rect: TRect): Boolean;
begin
  Result := (Point.X >= Rect.Left) and (Point.X <= Rect.Right) and
    (Point.Y >= Rect.Top) and (Point.Y <= Rect.Bottom);
end;

// =============================================================================

// Swaps red and blue values

function RGBGR(Value: TColor): TRGB4; overload;
asm
  BSWAP	EAX
  ROL	EAX, 24
end;

function RGBGR(Value: TRGB4): TColor; overload;
asm
  BSWAP	EAX
  ROL	EAX, 24
end;

// =============================================================================

function ReverseWord(W: Word): Word;
asm
  XCHG   AL, AH
end;

// =============================================================================

function Color2Float(Value: TColor): TFloatColor;
begin
  with Result do
  begin
    fRed := Value and $FF;
    fGreen := (Value and $FF00) shr 8;
    fBlue := (Value and $FF0000) shr 16;
  end;
end;

// =============================================================================

function RGB2HSV(Color: TRGB4): TRGB4;
// Convert RGB to HSVI
var
  R, G, B: Integer;
  Min, Max, Delta: Integer;
begin
  R := Color.rgbRed;
  G := Color.rgbGreen;
  B := Color.rgbBlue;
  Min := MinIntValue([R, G, B]);
  Max := MaxIntValue([R, G, B]);
  Delta := Max - Min;
  Result.Color := 0;
  Result.hsvValue := Max;
  Result.hsvIntensity := (R + G + B) div 3;
  if Max <> 0 then
    Result.hsvSaturation := MulDiv(Delta, 255, Max);
  if Result.hsvSaturation <> 0 then
  begin
    if R = Max then
      Result.hsvHue := MulDiv(G - B, 85, 2 * Delta)
    else if G = Max then
      Result.hsvHue := 85 + MulDiv(B - R, 85, 2 * Delta)
    else if B = Max then
      Result.hsvHue := 170 + MulDiv(R - G, 85, 2 * Delta)
  end;
end;

// =============================================================================

// TRGB4 color to gray value as integer

function ColorToGray(Col: TRGB4): Cardinal;
begin
  Result := (Col.rgbRed * 30 + Col.rgbGreen * 59 + Col.rgbBlue * 11) div 100;
end;

// =============================================================================

// Gray value to TRGB4 color

function GrayToColor(Color: Cardinal): TRGB4;
begin
  Result.Color := Color + Color shl 8 + Color shl 16;
end;

// =============================================================================

// Return mathematical difference between two TRGB4 colors

function ColorDiff(var Col1, Col2: TRGB4): Integer;
begin
  Result := ((Col1.rgbRed - Col2.rgbRed) * (Col1.rgbRed - Col2.rgbRed)) +
    ((Col1.rgbGreen - Col2.rgbGreen) * (Col1.rgbGreen - Col2.rgbGreen)) +
    ((Col1.rgbBlue - Col2.rgbBlue) * (Col1.rgbBlue - Col2.rgbBlue));
end;

{ ************************************************************************* }

function LeftS(const Txt: string; Size: Integer): string;
{ Returns <Size> chars from left of <Txt> }
begin
  if Size > 0 then
    Result := Copy(Txt, 1, Size)
  else
    Result := '';
end;

{ ************************************************************************* }

function EndS(const Txt: string; Pos: Integer): string;
{ Returns all chars from <Pos> to end of <Txt> }
begin
  if (Pos <= Length(Txt)) then
    Result := Copy(Txt, Pos, Length(Txt) - Pos + 1)
  else
    Result := '';
end;

{ ************************************************************************* }

function Subst(Txt: string; const S1, S2: string): string;
{ Substitute <S2> for <S1> throughout string }
{ S2 can contain S1 }
var
  P: Integer;
begin
  Result := Txt;
  if (S1 = S2) or (Length(Txt) < Length(S1)) then
    Exit;
  Result := '';
  repeat
    P := Pos(S1, Txt);
    if P <> 0 then
    begin
      Result := Result + LeftS(Txt, P - 1) + S2;
      Txt := EndS(Txt, P + Length(S1));
    end
    else
    begin
      Result := Result + Txt;
    end;
  until P = 0;
end;

// =============================================================================

// TPhotoBitmap

// =============================================================================

constructor TPhotoBitmap.Create;
begin
  inherited Create;
  SL := TStringList.Create;
  FCanvas := TCanvas.Create;
  FWidth := 1;
  FHeight := 1;
  FVisible := True;
  FDC := CreateCompatibleDC(0);
  SetStretchBltMode(FDC, STRETCH_DELETESCANS);
  New(FBMPInfo);
  FillChar(FBMPInfo^, SizeOf(TBitmapInfo), 0);
  with FBMPInfo^.bmiHeader do
  begin
    biSize := SizeOf(TBitmapInfoHeader);
    biPlanes := 1;
    biBitCount := 32;
    biCompression := BI_RGB;
  end;
  GetMem(FCalcs, 0);
  GetMem(FFCalcs, 0);
  NewDIBSection;
end;

// =============================================================================

destructor TPhotoBitmap.Destroy;
{ Clean up }
begin
  if (FFloatColors <> nil) then
    FreeMem(FFloatColors);
  FreeMem(FCalcs);
  FreeMem(FFCalcs);
  if (FSERFile <> nil) then
    FSERFile.Free;
  if OldBitmap <> 0 then
    SelectObject(FDC, OldBitmap);
  if FHandle <> 0 then
    DeleteObject(FHandle);
  FCanvas.Free;
  DeleteDC(FDC);
  Dispose(FBMPInfo);
  SL.Free;
  inherited Destroy;
end;

// =============================================================================

procedure TPhotoBitmap.NewDIBSection;
// Create a new DIB section
var
  I: Integer;
begin
  if OldBitmap <> 0 then
    SelectObject(FDC, OldBitmap);
  // Delete old DIBSection
  if FHandle <> 0 then
    DeleteObject(FHandle);
  FPixelCount := FWidth * FHeight;
  FRowInc := FWidth * SizeOf(TRGB4);
  FFRowInc := FWidth * SizeOf(TFloatColor);
  FSize := FHeight * FWidth * SizeOf(TRGB4);
  FFSize := FHeight * FWidth * SizeOf(TFloatColor);
  FBMPInfo^.bmiHeader.biHeight := -FHeight;
  FBMPInfo^.bmiHeader.biWidth := FWidth;
  // Create the new DIB section
  GDIFlush;
  FHandle := CreateDIBSection(FDC, FBMPInfo^, DIB_RGB_COLORS, FBits, 0, 0);
  OldBitmap := SelectObject(FDC, FHandle);
  // Assign FDC to canvas
  FCanvas.Handle := FDC;
  if (FFloatColors <> nil) then
    FreeMem(FFloatColors);
  FFloatColors := nil;
  if not FNoFloatColors then
  begin
    GetMem(FFloatColors, FFSize);
    ClearF;
  end;
  // Precalculate scan line offsets
  FreeMem(FCalcs);
  FreeMem(FFCalcs);
  GetMem(FCalcs, FHeight * SizeOf(Integer));
  GetMem(FFCalcs, FHeight * SizeOf(Integer));
  for I := 0 to FHeight - 1 do
  begin
    FCalcs^[I] := Integer(FBits) + (I * FRowInc);
    FFCalcs^[I] := Integer(FFloatColors) + (I * FFRowInc);
  end;
  Fill(FColor);
  if (FSERFile <> nil) then
    FSERFile.Free;
  FSERFile := TSERFile.Create;
  with FSERFile do
  begin
    ImageWidth := FWidth;
    ImageHeight := FHeight;
    if FMonochrome then
      ColorID := SER_MONO
    else
      ColorID := SER_RGB;
    PixelDepth := FPixelDepth;
  end;
  FBMPChanged := True;
end;

// =============================================================================

procedure TPhotoBitmap.Clear;
// Clear bitmap
begin
  ZeroMemory(FBits, FSize);
end;

// =============================================================================

procedure TPhotoBitmap.ClearF;
// Clear floatcolors
begin
  if (FFloatColors <> nil) then
    ZeroMemory(FFloatColors, FFSize);
end;

// =============================================================================

procedure TPhotoBitmap.SetHeight(Value: Integer);
begin
  if (Value <> FHeight) then
  begin
    FHeight := MaxOf(Value, 1);
    NewDIBSection;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.SetWidth(Value: Integer);
begin
  if (Value <> FWidth) then
  begin
    FWidth := MaxOf(Value, 1);
    NewDIBSection;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.SetSize(WD, HT: Integer);
begin
  FWidth := MaxOf(WD, 1);
  FHeight := MaxOf(HT, 1);
  NewDIBSection;
end;

// =============================================================================

// Set pixel depth in bits (= white level)
procedure TPhotoBitmap.SetPixelDepth(Value: Integer);
begin
  if (FPixelDepth <> Value) then
  begin
    FPixelDepth := Value;
    NewDIBSection;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.Fill(Value: TColor);
var
  I: Integer;
  Col: TRGB4;
begin
  Col := RGBGR(Value);
  if FMonochrome then
  begin
    Col.rgbGreen := Col.rgbRed;
    Col.rgbBlue := Col.rgbRed;
  end;
  for I := 0 to FPixelCount - 1 do
    PRGB4(Integer(FBits) + I * SizeOf(TRGB4))^ := Col;
end;

// =============================================================================

procedure TPhotoBitmap.FillTransparent;
var
  I: Integer;
  Col: TRGB4;
begin
  Col := RGBGR(FTransparentColor);
  for I := 0 to FPixelCount - 1 do
    PRGB4(Integer(FBits) + I * SizeOf(TRGB4))^ := Col;
end;

// =============================================================================

// Fast pixel and scanline access

procedure TPhotoBitmap.SetPixel(X, Y: Integer; Clr: TColor);
begin
  PRGB4(FCalcs^[Y] + (X * SizeOf(TRGB4)))^ := RGBGR(Clr);
end;

// =============================================================================

function TPhotoBitmap.GetPixel(X, Y: Integer): TColor;
begin
  Result := RGBGR(PRGB4(FCalcs^[Y] + (X * SizeOf(TRGB4)))^);
end;

// =============================================================================

procedure TPhotoBitmap.SetLine(Y: Integer; Line: Pointer);
begin
  CopyMemory(Pointer(FCalcs^[Y]), Line, FRowInc);
end;

// =============================================================================

function TPhotoBitmap.GetLine(Y: Integer): Pointer;
begin
  Result := Pointer(FCalcs^[Y]);
end;

// =============================================================================

// Fast float pixel and scanline access

procedure TPhotoBitmap.SetPixelF(X, Y: Integer; Clr: TFloatColor);
begin
  PFloatColor(FFCalcs^[Y] + (X * SizeOf(TFloatColor)))^ := Clr;
end;

// =============================================================================

function TPhotoBitmap.GetPixelF(X, Y: Integer): TFloatColor;
begin
  Result := PFloatColor(FFCalcs^[Y] + (X * SizeOf(TFloatColor)))^;
end;

// =============================================================================

procedure TPhotoBitmap.SetLineF(Y: Integer; Line: Pointer);
begin
  CopyMemory(Pointer(FFCalcs^[Y]), Line, FFRowInc);
end;

// =============================================================================

function TPhotoBitmap.GetLineF(Y: Integer): Pointer;
begin
  Result := Pointer(FFCalcs^[Y]);
end;

// =============================================================================

procedure TPhotoBitmap.FastLine(x1, y1, x2, y2: Integer);
// Fast line-drawing routine using Bresenham's Algorithm
var
  D, ax, ay, SX, SY, dx, dy: Integer;
  Col: TRGB4;
begin
  Col := RGBGR(Canvas.Pen.Color);
  if FMonochrome then
  begin
    Col.rgbGreen := Col.rgbRed;
    Col.rgbBlue := Col.rgbRed;
  end;
  dx := x2 - x1;
  ax := Abs(dx) shl 1;
  if dx < 0 then
    SX := -1
  else
    SX := 1;
  dy := y2 - y1;
  ay := Abs(dy) shl 1;
  if dy < 0 then
    SY := -1
  else
    SY := 1;
  if (x1 >= 0) and (x1 < FWidth) and (y1 >= 0) and (y1 < FHeight) then
    PRGB4(FCalcs^[y1] + (x1 * SizeOf(TRGB4)))^ := Col;
  if ax > ay then
  begin
    D := ay - (ax shr 1);
    while x1 <> x2 do
    begin
      if D > -1 then
      begin
        Inc(y1, SY);
        Dec(D, ax);
      end;
      Inc(x1, SX);
      Inc(D, ay);
      if (x1 >= 0) and (x1 < FWidth) and (y1 >= 0) and (y1 < FHeight) then
        PRGB4(FCalcs^[y1] + (x1 * SizeOf(TRGB4)))^ := Col;
    end;
  end
  else
  begin
    D := ax - (ay shr 1);
    while y1 <> y2 do
    begin
      if D >= 0 then
      begin
        Inc(x1, SX);
        Dec(D, ay);
      end;
      Inc(y1, SY);
      Inc(D, ax);
      if (x1 >= 0) and (x1 < FWidth) and (y1 >= 0) and (y1 < FHeight) then
        PRGB4(FCalcs^[y1] + (x1 * SizeOf(TRGB4)))^ := Col;
    end;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.Rotate(OX, OY: Integer; Angle: Single);
var
  Rdx, Rdy, Ddx, Ddy: Single;
  x1, x2, y1, y2: Integer;
  dx, dy, Rads: Single;
  Temp: PRGB4;

begin
  GetMem(Temp, FSize);
  try
    begin
      CopyMemory(Temp, FBits, FSize);
      ZeroMemory(FBits, FSize);
      { Convert degrees to radians }
      Rads := (0 - Angle) * Pi / 180;
      { Calculate the right vector }
      Rdx := Cos(Rads);
      Rdy := Sin(Rads);
      { Calculate the down vector }
      Ddx := Cos(Rads + Pi / 2);
      Ddy := Sin(Rads + Pi / 2);
      { Rotate about OX, OY }
      for y1 := 0 to FHeight - 1 do
      begin
        dx := OX + (Ddx * (y1 - OY));
        dy := OY + (Ddy * (y1 - OY));
        for x1 := 0 to FWidth - 1 do
        begin
          { Find corresponding pixel in the source bitmap }
          x2 := Round(Rdx * (x1 - OX) + dx);
          y2 := Round(Rdy * (x1 - OX) + dy);
          { Copy pixel color from source to destination }
          if (x2 >= 0) and (x2 < FWidth) and (y2 >= 0) and (y2 < FHeight) then
          begin
            PRGB4(FCalcs^[y1] + (x1 * SizeOf(TRGB4)))^ :=
              PRGB4(FCalcs^[y2] - Integer(FBits) + Integer(Temp) +
              (x2 * SizeOf(TRGB4)))^
          end;
        end;
      end;
    end;
  finally
    FreeMem(Temp);
  end;
end;

// =============================================================================

procedure TPhotoBitmap.SetColor(Value: TColor);
// Set background color
begin
  if (Value <> FColor) then
  begin
    Fill(Value);
    FColor := Value;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.SetMonochrome(Value: Boolean);
// Set monochrome status
begin
  if (Value <> FMonochrome) then
  begin
    FMonochrome := Value;
    NewDIBSection;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.FlipX;
// Flip image left/right
var
  X, Y, Z: Integer;
  Col: TRGB4;
begin
  for X := 0 to (FWidth div 2) - 1 do
  begin
    Z := FWidth - 1 - X;
    for Y := 0 to FHeight - 1 do
    begin
      Col := PRGB4(FCalcs^[Y] + (X * SizeOf(TRGB4)))^;
      PRGB4(FCalcs^[Y] + (X * SizeOf(TRGB4)))^ :=
        PRGB4(FCalcs^[Y] + (Z * SizeOf(TRGB4)))^;
      PRGB4(FCalcs^[Y] + (Z * SizeOf(TRGB4)))^ := Col;
    end;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.FlipY;
// Flip image top/bottom
var
  X, Y, Z: Integer;
  Col: TRGB4;
begin
  for Y := 0 to FHeight div 2 - 1 do
  begin
    Z := FHeight - 1 - Y;
    for X := 0 to FWidth - 1 do
    begin
      Col := PRGB4(FCalcs^[Y] + (X * SizeOf(TRGB4)))^;
      PRGB4(FCalcs^[Y] + (X * SizeOf(TRGB4)))^ :=
        PRGB4(FCalcs^[Z] + (X * SizeOf(TRGB4)))^;
      PRGB4(FCalcs^[Z] + (X * SizeOf(TRGB4)))^ := Col;
    end;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.Assign(Value: TPhotoBitmap);
begin
  if Assigned(Value) then
  begin
    FWidth := Value.Width;
    FHeight := Value.Height;
    FPixelDepth := Value.PixelDepth;
    FColor := Value.Color;
    FNoFloatColors := Value.NoFloatColors;
    FMonochrome := Value.Monochrome;
    NewDIBSection;
    FParent := Value.Parent;
    FVisible := Value.Visible;
    CopyMemory(FBits, Value.Bits, FSize);
    if not FNoFloatColors then
      CopyMemory(FFloatColors, Value.FloatColors,
        FHeight * FWidth * SizeOf(TFloatColor));
  end;
end;

// =============================================================================

procedure TPhotoBitmap.MakeCompatible(Dest: TPhotoBitmap);
begin
  if (Dest.Width <> Width) or (Dest.Height <> Height) or (Dest.Color <> Color)
    or (Dest.Monochrome <> Monochrome) or (Dest.PixelDepth <> PixelDepth) or
    (Dest.NoFloatColors <> NoFloatColors) then
    with Dest do
    begin
      FWidth := Width;
      FHeight := Height;
      FColor := Color;
      FPixelDepth := PixelDepth;
      FMonochrome := Monochrome;
      FNoFloatColors := NoFloatColors;
      NewDIBSection;
    end;
end;

// =============================================================================

function TPhotoBitmap.CreateCopy: TPhotoBitmap;
begin
  Result := TPhotoBitmap.Create;
  Result.FWidth := FWidth;
  Result.FHeight := FHeight;
  Result.FColor := FColor;
  Result.FPixelDepth := FPixelDepth;
  Result.FMonochrome := FMonochrome;
  Result.FNoFloatColors := FNoFloatColors;
  Result.NewDIBSection;
  Result.FVisible := Visible;
  CopyMemory(Result.FBits, FBits, FSize);
  if not FNoFloatColors then
    CopyMemory(Result.FFloatColors, FFloatColors, FFSize);
end;

// =============================================================================

procedure TPhotoBitmap.RadialCorrect(Amount: Single);
// Perform radial correction on the bitmap
var
  x1, x2, y1, y2: Integer;
  dx, dy: Integer;
  Radius: Single;
  Diag: Single;
  Dist: Single;
  Temp: PRGB4;
begin
  GetMem(Temp, FSize);
  try
    begin
      CopyMemory(Temp, FBits, FSize);
      ZeroMemory(FBits, FSize);
      Diag := (Sqrt(Sqr(FWidth) + Sqr(FHeight))) / 2;
      for y1 := 0 to FHeight - 1 do
      begin
        for x1 := 0 to FWidth - 1 do
        begin
          dx := x1 - (FWidth div 2);
          dy := y1 - (FHeight div 2);
          Radius := Sqrt(Sqr(dx) + Sqr(dy));
          Dist := 1 - (Radius / Diag);
          x2 := x1 + Round(Amount * Dist * dx);
          y2 := y1 + Round(Amount * Dist * dy);
          if (x2 >= 0) and (x2 < FWidth) and (y2 >= 0) and (y2 < FHeight) then
          begin
            PRGB4(FCalcs^[y1] + (x1 * SizeOf(TRGB4)))^ :=
              PRGB4(FCalcs^[y2] - Integer(FBits) + Integer(Temp) +
              (x2 * SizeOf(TRGB4)))^;
          end;
        end;
      end;
    end;
  finally
    FreeMem(Temp);
  end;
end;

// =============================================================================

procedure TPhotoBitmap.RadialCorrect2(K: Single);
// Barrel/Pincushion correction
// Bicubic interpolation
// K : Barrel/Pincushion correction
// K > 0 corrects for pincushion
// K < 0 corrects for barrel

var
  BM: TPhotoBitmap;
  Col: TRGB4;
  xpx, ypx, gmx, gmy: Integer;
  x1, y1, x9, y9: Integer;
  x5, y5, x8, y8: Single;
  Kr, Rd1, k1, k2: Single;
  r00, r10, r20, r30, r01, r11, r21, r31, r02, r12, r22, r32, r03, r13, r23,
    r33: Integer;
  g00, g10, g20, g30, g01, g11, g21, g31, g02, g12, g22, g32, g03, g13, g23,
    g33: Integer;
  b00, b10, b20, b30, b01, b11, b21, b31, b02, b12, b22, b32, b03, b13, b23,
    b33: Integer;
  a1, a2, a3, b1, b2, b3, c1, c2: Single;
  dx, dy, dx1, dy1, dx2, dx3, dy2, dy3: Single;
  ra, rb, RC, rd: Single;
  ga, gb, gc, GD: Single;
  ba, bb, BC, bd: Single;
  PRGB: Integer;

const
  k16: Single = 1 / 6;
  k13: Single = 1 / 3;
  k12: Single = 1 / 2;

  procedure Par3(c0, c1, c2, c3: Integer; dx, dx2, dx3: Single;
    var cj: Single); inline;
  var
    d0, D2, d3: Integer;
    a1, a2, a3: Single;
  begin
    d0 := c0 - c1;
    D2 := c2 - c1;
    d3 := c3 - c1;
    a1 := D2 - (k13 * d0) - (k16 * d3);
    a2 := k12 * (d0 + D2);
    a3 := k16 * (d3 - d0) - (k12 * D2);
    cj := c1 + (a1 * dx) + (a2 * dx2) + (a3 * dx3);
  end;

  procedure Pbr3(c0, c1, c2, c3, dy, dy2, dy3: Single; var cj: Integer); inline;
  var
    d0, D2, d3: Single;
    a1, a2, a3: Single;
  begin
    d0 := c0 - c1;
    D2 := c2 - c1;
    d3 := c3 - c1;
    a1 := D2 - (k13 * d0) - (k16 * d3);
    a2 := k12 * (d0 + D2);
    a3 := k16 * (d3 - d0) - (k12 * D2);
    cj := Round(c1 + (a1 * dy) + (a2 * dy2) + (a3 * dy3));
    if (cj > 255) then
      cj := 255;
  end;

begin
  BM := TPhotoBitmap.Create;
  try
    BM.Width := Width + 2;
    BM.Height := Height + 2;
    CopyRectPixels(BM, 0, 0, 0, 0, Width, Height);
    Fill(clWhite);
    dx1 := Width / 2;
    dy1 := Height / 2; { Center }
    k1 := 1 - K;
    k2 := K / Sqr(dx1);
    { Extrapolation }
    for x1 := 0 to Width - 1 do
      BM.Pixels[x1, Height] := BM.Pixels[x1, Height - 1];
    for x1 := 0 to Width - 1 do
      BM.Pixels[x1, Height + 1] := BM.Pixels[x1, Height - 1];
    for y1 := 0 to Height + 1 do
      BM.Pixels[Width, y1] := BM.Pixels[Width - 1, y1];
    for y1 := 0 to Height + 1 do
      BM.Pixels[Width + 1, y1] := BM.Pixels[Width - 1, y1];
    for y1 := 0 to Height - 1 do
    begin
      for x1 := 0 to Width - 1 do
      begin
        x5 := x1 - dx1;
        y5 := y1 - dy1;
        Rd1 := Sqr(x5) + Sqr(y5);
        Kr := k1 + k2 * Rd1;
        x8 := dx1 + Kr * x5;
        y8 := dy1 + Kr * y5;
        x9 := Trunc(x8); { Reference }
        y9 := Trunc(y8); { in Source Framebuffer }
        If (x9 > 0) And (x9 < Width) And (y9 > 0) And (y9 < Height) then
        begin
          dx := x8 - x9; { Difference }
          dy := y8 - y9;
          dx2 := Sqr(dx);
          dx3 := dx * dx2;
          dy2 := Sqr(dy);
          dy3 := dy * dy2;
          Dec(y9); { -1 }
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 - 1) * SizeOf(TRGB4)))^;
          r00 := Col.bgrRed;
          g00 := Col.bgrGreen;
          b00 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + (x9 * SizeOf(TRGB4)))^;
          r10 := Col.bgrRed;
          g10 := Col.bgrGreen;
          b10 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 + 1) * SizeOf(TRGB4)))^;
          r20 := Col.bgrRed;
          g20 := Col.bgrGreen;
          b20 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 + 2) * SizeOf(TRGB4)))^;
          r30 := Col.bgrRed;
          g30 := Col.bgrGreen;
          b30 := Col.bgrBlue;
          Inc(y9); { 0 }
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 - 1) * SizeOf(TRGB4)))^;
          r01 := Col.bgrRed;
          g01 := Col.bgrGreen;
          b01 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + (x9 * SizeOf(TRGB4)))^;
          r11 := Col.bgrRed;
          g11 := Col.bgrGreen;
          b11 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 + 1) * SizeOf(TRGB4)))^;
          r21 := Col.bgrRed;
          g21 := Col.bgrGreen;
          b21 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 + 2) * SizeOf(TRGB4)))^;
          r31 := Col.bgrRed;
          g31 := Col.bgrGreen;
          b31 := Col.bgrBlue;
          Inc(y9); { +1 }
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 - 1) * SizeOf(TRGB4)))^;
          r02 := Col.bgrRed;
          g02 := Col.bgrGreen;
          b02 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + (x9 * SizeOf(TRGB4)))^;
          r12 := Col.bgrRed;
          g12 := Col.bgrGreen;
          b12 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 + 1) * SizeOf(TRGB4)))^;
          r22 := Col.bgrRed;
          g22 := Col.bgrGreen;
          b22 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 + 2) * SizeOf(TRGB4)))^;
          r32 := Col.bgrRed;
          g32 := Col.bgrGreen;
          b32 := Col.bgrBlue;
          Inc(y9); { +2 }
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 - 1) * SizeOf(TRGB4)))^;
          r03 := Col.bgrRed;
          g03 := Col.bgrGreen;
          b03 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + (x9 * SizeOf(TRGB4)))^;
          r13 := Col.bgrRed;
          g13 := Col.bgrGreen;
          b13 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 + 1) * SizeOf(TRGB4)))^;
          r23 := Col.bgrRed;
          g23 := Col.bgrGreen;
          b23 := Col.bgrBlue;
          Col := PRGB4(BM.FCalcs^[y9] + ((x9 + 2) * SizeOf(TRGB4)))^;
          r33 := Col.bgrRed;
          g33 := Col.bgrGreen;
          b33 := Col.bgrBlue;
          Par3(r00, r10, r20, r30, dx, dx2, dx3, ra);
          Par3(r01, r11, r21, r31, dx, dx2, dx3, rb);
          Par3(r02, r12, r22, r32, dx, dx2, dx3, RC);
          Par3(r03, r13, r23, r33, dx, dx2, dx3, rd);
          Pbr3(ra, rb, RC, rd, dy, dy2, dy3, r11);
          Col.bgrRed := r11;
          if not FMonochrome then
          begin
            Par3(g00, g10, g20, g30, dx, dx2, dx3, ga);
            Par3(g01, g11, g21, g31, dx, dx2, dx3, gb);
            Par3(g02, g12, g22, g32, dx, dx2, dx3, gc);
            Par3(g03, g13, g23, g33, dx, dx2, dx3, GD);
            Pbr3(ga, gb, gc, GD, dy, dy2, dy3, g11);
            Par3(b00, b10, b20, b30, dx, dx2, dx3, ba);
            Par3(b01, b11, b21, b31, dx, dx2, dx3, bb);
            Par3(b02, b12, b22, b32, dx, dx2, dx3, BC);
            Par3(b03, b13, b23, b33, dx, dx2, dx3, bd);
            Pbr3(ba, bb, BC, bd, dy, dy2, dy3, b11);
            Col.bgrGreen := g11;
            Col.bgrBlue := b11;
          end
          else
          begin
            Col.bgrGreen := Col.bgrRed;
            Col.bgrBlue := Col.bgrRed;
          end;
          PRGB4(FCalcs^[y1] + (x1 * SizeOf(TRGB4)))^ := Col;
        end; { If... }
      end; { x1... }
    end; { y1... }
  finally
    BM.Free;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.CopyRectPixels(Dst: TPhotoBitmap;
  DstX, DstY, SrcX, SrcY, W, H: Integer);
// Copy a rectangle of pixels from one PhotoBitmap to another
var
  X, Y, WD, HT: Integer;
begin
  // If same rectangles and dibsections, just copy all the bits
  if (DstX = SrcX) and (DstY = SrcY) and (W = FWidth) and (W = Dst.FWidth) and
    (H = FHeight) and (H = Dst.FHeight) then
    CopyMemory(Dst.FBits, FBits, FSize)
  else
  // Copy the largest acceptable rectangle
  begin
    WD := MinOf(W, FWidth - SrcX);
    WD := MinOf(WD, Dst.Width - DstX);
    HT := MinOf(H, FHeight - SrcY);
    HT := MinOf(HT, Dst.Height - DstY);
    for Y := 0 to HT - 1 do
      for X := 0 to WD - 1 do
        Dst.Pixels[DstX + X, DstY + Y] := Pixels[SrcX + X, SrcY + Y];
  end;
end;

// =============================================================================

procedure TPhotoBitmap.ConvolveF(Dims: Integer; Kernel: array of Single;
  Scale: Single);
// do Dims x Dims matrix convolution, using Kernel values
// Multiply each result by Scale
// Perform on FloatColors
var
  I, J, K, X, Y: Integer;
  XS, YS: Integer;
  NewR, NewG, NewB: Single;
  Radius: Integer;
  Col: TFloatColor;
  Ptr: PFloatColor;
  Dst: PFloatColor;
begin
  Dst := AllocMem(FFSize);
  Radius := Dims div 2;
  for Y := 0 to Height - 1 do
  begin
    for X := 0 to FWidth - 1 do
    begin
      K := 0;
      NewR := 0;
      NewG := 0;
      NewB := 0;
      for J := (Y - Radius) to (Y + Radius) do
      begin
        if (J < 0) then
          YS := 0
        else if (J >= FHeight) then
          YS := FHeight - 1
        else
          YS := J;
        for I := (X - Radius) to (X + Radius) do
        begin
          if (I < 0) then
            XS := 0
          else if (I >= FWidth) then
            XS := FWidth - 1
          else
            XS := I;
          Ptr := PFloatColor(FFCalcs^[YS] + XS * SizeOf(TFloatColor));
          if Kernel[K] <> 0 then
          begin
            NewR := NewR + (Ptr^.fRed * Kernel[K]);
            if not FMonochrome then
            begin
              NewG := NewG + (Ptr^.fGreen * Kernel[K]);
              NewB := NewB + (Ptr^.fBlue * Kernel[K]);
            end;
          end;
          Inc(K);
        end;
      end;
      if Scale = 1 then
      begin
        Col.fRed := NewR;
        if not FMonochrome then
        begin
          Col.fGreen := NewG;
          Col.fBlue := NewB;
        end;
      end
      else
      begin
        Col.fRed := NewR * Scale;
        if not FMonochrome then
        begin
          Col.fGreen := NewG * Scale;
          Col.fBlue := NewB * Scale;
        end;
      end;
      if FMonochrome then
      begin
        Col.fGreen := Col.fRed;
        Col.fBlue := Col.fRed;
      end;
      PFloatColor(Integer(Dst) + (Y * FFRowInc) + (X * SizeOf(TFloatColor)
        ))^ := Col;
    end;
  end;
  CopyMemory(FFloatColors, Dst, FFSize);
  FreeMem(Dst);
end;

// =============================================================================

procedure TPhotoBitmap.DespeckleF;
// Fast median filter on the floatcolors data
var
  X, Y: Integer;
  Floats: PFloatColor;
  Src, Dst: PFloatColor;
  Col1, Col2, Col3: TFloatColor;
  a1, b1, c1, D1, E1, F1, G1, H1, i1: TFloatColor;
  a2, b2, c2, D2, E2, F2, G2, H2, i2: TFloatColor;
  a3, b3, c3, d3, E3, F3, G3, H3, i3: TFloatColor;
  Tmp: TFloatColor;
  Clr: Single;
begin
  Floats := AllocMem(FFSize);
  try
    // Red bytes
    for Y := 0 to FHeight - 1 do
    begin
      Src := PFloatColor(Integer(FFloatColors) + SizeOf(TFloatColor) *
        (FWidth * Y + 1));
      Dst := PFloatColor(Integer(Floats) + SizeOf(TFloatColor) *
        (FWidth * Y + 1));
      Dec(Src, FWidth + 1);
      for X := 0 to FWidth - 1 do
      begin
        if (Y = 0) or (Y = FHeight - 1) or (X = 0) or (X = FWidth - 1) then
        begin
          Tmp := PFloatColor(Integer(FloatColors) + (FFRowInc * Y) +
            (X * SizeOf(TFloatColor)))^;
          if FMonochrome then
          begin
            Tmp.fGreen := Tmp.fRed;
            Tmp.fBlue := Tmp.fRed;
          end;
          PFloatColor(Integer(Floats) + (FFRowInc * Y) +
            (X * SizeOf(TFloatColor)))^ := Tmp;
          Continue;
        end;
        a1 := Src^;
        a2 := a1;
        a3 := a1;
        Inc(Src);
        b1 := Src^;
        b2 := b1;
        b3 := b1;
        Inc(Src);
        c1 := Src^;
        c2 := c1;
        c3 := c1;
        Inc(Src, FWidth - 2);
        D1 := Src^;
        D2 := D1;
        d3 := D1;
        Inc(Src);
        E1 := Src^;
        E2 := E1;
        E3 := E1;
        Inc(Src);
        F1 := Src^;
        F2 := F1;
        F3 := F1;
        Inc(Src, FWidth - 2);
        G1 := Src^;
        G2 := G1;
        G3 := G1;
        Inc(Src);
        H1 := Src^;
        H2 := H1;
        H3 := H1;
        Inc(Src);
        i1 := Src^;
        i2 := i1;
        i3 := i1;
        Dec(Src, 2 * FWidth + 1);
        if b1.fRed > c1.fRed then
        begin
          Tmp := b1;
          b1 := c1;
          c1 := Tmp;
        end;
        if E1.fRed > F1.fRed then
        begin
          Tmp := E1;
          E1 := F1;
          F1 := Tmp;
        end;
        if H1.fRed > i1.fRed then
        begin
          Tmp := H1;
          H1 := i1;
          i1 := Tmp;
        end;
        if a1.fRed > b1.fRed then
        begin
          Tmp := a1;
          a1 := b1;
          b1 := Tmp;
        end;
        if D1.fRed > E1.fRed then
        begin
          Tmp := D1;
          D1 := E1;
          E1 := Tmp;
        end;
        if G1.fRed > H1.fRed then
        begin
          Tmp := G1;
          G1 := H1;
          H1 := Tmp;
        end;
        if b1.fRed > c1.fRed then
        begin
          Tmp := b1;
          b1 := c1;
          c1 := Tmp;
        end;
        if E1.fRed > F1.fRed then
        begin
          Tmp := E1;
          E1 := F1;
          F1 := Tmp;
        end;
        if H1.fRed > i1.fRed then
        begin
          Tmp := H1;
          H1 := i1;
          i1 := Tmp;
        end;
        if a1.fRed > D1.fRed then
          D1 := a1;
        if F1.fRed > i1.fRed then
          F1 := i1;
        if E1.fRed > H1.fRed then
        begin
          Tmp := E1;
          E1 := H1;
          H1 := Tmp;
        end;
        if D1.fRed > G1.fRed then
          G1 := D1;
        if b1.fRed > E1.fRed then
          E1 := b1;
        if c1.fRed > F1.fRed then
          c1 := F1;
        if E1.fRed > H1.fRed then
          E1 := H1;
        if E1.fRed > c1.fRed then
        begin
          Tmp := E1;
          E1 := c1;
          c1 := Tmp;
        end;
        if G1.fRed > E1.fRed then
          E1 := G1;
        if E1.fRed > c1.fRed then
          Col1 := c1
        else
          Col1 := E1;

        if not FMonochrome then
        begin
          if b2.fGreen > c2.fGreen then
          begin
            Tmp := b2;
            b2 := c2;
            c2 := Tmp;
          end;
          if E2.fGreen > F2.fGreen then
          begin
            Tmp := E2;
            E2 := F2;
            F2 := Tmp;
          end;
          if H2.fGreen > i2.fGreen then
          begin
            Tmp := H2;
            H2 := i2;
            i2 := Tmp;
          end;
          if a2.fGreen > b2.fGreen then
          begin
            Tmp := a2;
            a2 := b2;
            b2 := Tmp;
          end;
          if D2.fGreen > E2.fGreen then
          begin
            Tmp := D2;
            D2 := E2;
            E2 := Tmp;
          end;
          if G2.fGreen > H2.fGreen then
          begin
            Tmp := G2;
            G2 := H2;
            H2 := Tmp;
          end;
          if b2.fGreen > c2.fGreen then
          begin
            Tmp := b2;
            b2 := c2;
            c2 := Tmp;
          end;
          if E2.fGreen > F2.fGreen then
          begin
            Tmp := E2;
            E2 := F2;
            F2 := Tmp;
          end;
          if H2.fGreen > i2.fGreen then
          begin
            Tmp := H2;
            H2 := i2;
            i2 := Tmp;
          end;
          if a2.fGreen > D2.fGreen then
            D2 := a2;
          if F2.fGreen > i2.fGreen then
            F2 := i2;
          if E2.fGreen > H2.fGreen then
          begin
            Tmp := E2;
            E2 := H2;
            H2 := Tmp;
          end;
          if D2.fGreen > G2.fGreen then
            G2 := D2;
          if b2.fGreen > E2.fGreen then
            E2 := b2;
          if c2.fGreen > F2.fGreen then
            c2 := F2;
          if E2.fGreen > H2.fGreen then
            E2 := H2;
          if E2.fGreen > c2.fGreen then
          begin
            Tmp := E2;
            E2 := c2;
            c2 := Tmp;
          end;
          if G2.fGreen > E2.fGreen then
            E2 := G2;
          if E2.fGreen > c2.fGreen then
            Col2 := c2
          else
            Col2 := E2;

          if b3.fBlue > c3.fBlue then
          begin
            Tmp := b3;
            b3 := c3;
            c3 := Tmp;
          end;
          if E3.fBlue > F3.fBlue then
          begin
            Tmp := E3;
            E3 := F3;
            F3 := Tmp;
          end;
          if H3.fBlue > i3.fBlue then
          begin
            Tmp := H3;
            H3 := i3;
            i3 := Tmp;
          end;
          if a3.fBlue > b3.fBlue then
          begin
            Tmp := a3;
            a3 := b3;
            b3 := Tmp;
          end;
          if d3.fBlue > E3.fBlue then
          begin
            Tmp := d3;
            d3 := E3;
            E3 := Tmp;
          end;
          if G3.fBlue > H3.fBlue then
          begin
            Tmp := G3;
            G3 := H3;
            H3 := Tmp;
          end;
          if b3.fBlue > c3.fBlue then
          begin
            Tmp := b3;
            b3 := c3;
            c3 := Tmp;
          end;
          if E3.fBlue > F3.fBlue then
          begin
            Tmp := E3;
            E3 := F3;
            F3 := Tmp;
          end;
          if H3.fBlue > i3.fBlue then
          begin
            Tmp := H3;
            H3 := i3;
            i3 := Tmp;
          end;
          if a3.fBlue > d3.fBlue then
            d3 := a3;
          if F3.fBlue > i3.fBlue then
            F3 := i3;
          if E3.fBlue > H3.fBlue then
          begin
            Tmp := E3;
            E3 := H3;
            H3 := Tmp;
          end;
          if d3.fBlue > G3.fBlue then
            G3 := d3;
          if b3.fBlue > E3.fBlue then
            E3 := b3;
          if c3.fBlue > F3.fBlue then
            c3 := F3;
          if E3.fBlue > H3.fBlue then
            E3 := H3;
          if E3.fBlue > c3.fBlue then
          begin
            Tmp := E3;
            E3 := c3;
            c3 := Tmp;
          end;
          if G3.fBlue > E3.fBlue then
            E3 := G3;
          if E3.fBlue > c3.fBlue then
            Col3 := c3
          else
            Col3 := E3;
          Dst^.fRed := (Col1.fRed + Col2.fRed + Col3.fRed) / 3;
          Dst^.fGreen := (Col1.fGreen + Col2.fGreen + Col3.fGreen) / 3;
          Dst^.fBlue := (Col1.fBlue + Col2.fBlue + Col3.fBlue) / 3;
        end
        else
        begin
          Dst^.fRed := Col1.fRed;
          Dst^.fGreen := Col1.fRed;
          Dst^.fBlue := Col1.fRed;
        end;
        Inc(Dst);
      end;
    end;
    // Copy to destination
    CopyMemory(FFloatColors, Floats, FFSize);
  finally
    FreeMem(Floats);
  end;
end;

// =============================================================================

procedure TPhotoBitmap.ColorDifference(Dst: TPhotoBitmap);
// Edge detection using color difference
var
  X, Y: Integer;
  BMP: Pointer;
  Src, Dest: PRGB4;
  A, B, C, D, E, F, G, H, I: TRGB4;
  Diff: Byte;
begin
  if Dst.Size = FSize then
  begin
    BMP := AllocMem(FSize);
    try
      // Red bytes
      for Y := 1 to FHeight - 2 do
      begin
        Src := PRGB4(Integer(FBits) + SizeOf(Integer) * (FWidth * Y + 1));
        Dest := PRGB4(Integer(BMP) + SizeOf(Integer) * (FWidth * Y + 1));
        Dec(Src, FWidth + 1);
        for X := 1 to FWidth - 2 do
        begin
          A := Src^;
          Inc(Src);
          B := Src^;
          Inc(Src);
          C := Src^;
          Inc(Src, FWidth - 2);
          D := Src^;
          Inc(Src);
          E := Src^;
          Inc(Src);
          F := Src^;
          Inc(Src, FWidth - 2);
          G := Src^;
          Inc(Src);
          H := Src^;
          Inc(Src);
          I := Src^;
          Dec(Src, 2 * FWidth + 1);
          // Calculate color difference for 8 surrounding pixels
          Diff := ClampByte
            (Round(Sqrt(((E.rgbRed - A.rgbRed) * (E.rgbRed - A.rgbRed)) +
            ((E.rgbGreen - A.rgbGreen) * (E.rgbGreen - A.rgbGreen)) +
            ((E.rgbBlue - A.rgbBlue) * (E.rgbBlue - A.rgbBlue)) +
            ((E.rgbRed - B.rgbRed) * (E.rgbRed - B.rgbRed)) +
            ((E.rgbGreen - B.rgbGreen) * (E.rgbGreen - B.rgbGreen)) +
            ((E.rgbBlue - B.rgbBlue) * (E.rgbBlue - B.rgbBlue)) +
            ((E.rgbRed - C.rgbRed) * (E.rgbRed - C.rgbRed)) +
            ((E.rgbGreen - C.rgbGreen) * (E.rgbGreen - C.rgbGreen)) +
            ((E.rgbBlue - C.rgbBlue) * (E.rgbBlue - C.rgbBlue)) +
            ((E.rgbRed - D.rgbRed) * (E.rgbRed - D.rgbRed)) +
            ((E.rgbGreen - D.rgbGreen) * (E.rgbGreen - D.rgbGreen)) +
            ((E.rgbBlue - D.rgbBlue) * (E.rgbBlue - D.rgbBlue)) +
            ((E.rgbRed - F.rgbRed) * (E.rgbRed - F.rgbRed)) +
            ((E.rgbGreen - F.rgbGreen) * (E.rgbGreen - F.rgbGreen)) +
            ((E.rgbBlue - F.rgbBlue) * (E.rgbBlue - F.rgbBlue)) +
            ((E.rgbRed - G.rgbRed) * (E.rgbRed - G.rgbRed)) +
            ((E.rgbGreen - G.rgbGreen) * (E.rgbGreen - G.rgbGreen)) +
            ((E.rgbBlue - G.rgbBlue) * (E.rgbBlue - G.rgbBlue)) +
            ((E.rgbRed - H.rgbRed) * (E.rgbRed - H.rgbRed)) +
            ((E.rgbGreen - H.rgbGreen) * (E.rgbGreen - H.rgbGreen)) +
            ((E.rgbBlue - H.rgbBlue) * (E.rgbBlue - H.rgbBlue)) +
            ((E.rgbRed - I.rgbRed) * (E.rgbRed - I.rgbRed)) +
            ((E.rgbGreen - I.rgbGreen) * (E.rgbGreen - I.rgbGreen)) +
            ((E.rgbBlue - I.rgbBlue) * (E.rgbBlue - I.rgbBlue)))));
          if Dst.Monochrome then
          begin
            Dest^.rgbRed := Diff;
            Dest^.rgbGreen := Diff;
            Dest^.rgbBlue := Diff;
          end
          else
          begin
            Dest^ := E;
            Dest^.rgbReserved := Diff;
          end;
          Inc(Dest);
        end;
      end;
      CopyMemory(Dst.Bits, BMP, FSize);
    finally
      FreeMem(BMP);
    end;
  end;
end;

// =============================================================================

function TPhotoBitmap.AverageFloatColors: TFloatColor;
// Get average of all pixels
var
  I: Integer;
  P: PFloatColor;
begin
  Result.fRed := 0;
  Result.fGreen := 0;
  Result.fBlue := 0;
  if (FFloatColors <> nil) then
  begin
    P := FFloatColors;
    for I := 0 to FPixelCount - 1 do
    begin
      Result.fRed := Result.fRed + P^.fRed;
      if not FMonochrome then
      begin
        Result.fGreen := Result.fGreen + P^.fGreen;
        Result.fBlue := Result.fBlue + P^.fBlue;
      end;
      Inc(P);
    end;
  end;
  Result.fRed := Result.fRed / FPixelCount;
  if not FMonochrome then
  begin
    Result.fGreen := Result.fGreen / FPixelCount;
    Result.fBlue := Result.fBlue / FPixelCount;
  end
  else
  begin
    Result.fGreen := Result.fRed;
    Result.fBlue := Result.fRed;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.FillFloatColors(Source: TPhotoBitmap;
  DstScale, SrcScale: Single);
// Add color values from source pixels to FloatColorArray
// Can scale amount added from source
// Dst = (Dst * DstScale) + (Srx * SrcScale)
var
  I: Integer;
  Src: PRGB4;
  Dst: PFloatColor;
begin
  if (FFloatColors <> nil) and (Source.Size = FSize) then
  begin
    Src := Source.Bits;
    Dst := FFloatColors;
    for I := 0 to FPixelCount - 1 do
    begin
      Dst^.fRed := (Dst^.fRed * DstScale) + (Src^.rgbRed * SrcScale);
      if not FMonochrome then
      begin
        Dst^.fGreen := (Dst^.fGreen * DstScale) + (Src^.rgbGreen * SrcScale);
        Dst^.fBlue := (Dst^.fBlue * DstScale) + (Src^.rgbBlue * SrcScale);
      end
      else
      begin
        Dst^.fGreen := Dst^.fRed;
        Dst^.fBlue := Dst^.fRed;
      end;
      Inc(Src);
      Inc(Dst);
    end;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.FillFloatFactors(Source: TPhotoBitmap;
  Value: TFloatColor; Mult: Single = 1);
// Fill FloatColors with scale factors for each pixel to make them equal Value
var
  I: Integer;
  Src: PFloatColor;
  Dst: PFloatColor;
begin
  if (FFloatColors <> nil) and (Source.Size = FSize) then
  begin
    Src := Source.FFloatColors;
    Dst := FFloatColors;
    for I := 0 to FPixelCount - 1 do
    begin
      if (Src^.fRed <> 0) then
        Dst^.fRed := Dst^.fRed + (Value.fRed / Src^.fRed)
      else
        Dst^.fRed := Dst^.fRed + 1;
      if not FMonochrome then
      begin
        if (Src^.fGreen <> 0) then
          Dst^.fGreen := Dst^.fGreen + (Value.fGreen / Src^.fGreen)
        else
          Dst^.fGreen := Dst^.fGreen + 1;
        if (Src^.fBlue <> 0) then
          Dst^.fBlue := Dst^.fBlue + (Value.fBlue / Src^.fBlue)
        else
          Dst^.fBlue := Dst^.fBlue + 1;
        if (Mult <> 1) then
        begin
          Dst^.fRed := Dst^.fRed * Mult;
          Dst^.fGreen := Dst^.fGreen * Mult;
          Dst^.fBlue := Dst^.fBlue * Mult;
        end;
      end
      else
      begin
        if (Mult <> 1) then
        begin
          Dst^.fRed := Dst^.fRed * Mult;
          Dst^.fGreen := Dst^.fRed;
          Dst^.fBlue := Dst^.fRed;
        end;
      end;
      Inc(Src);
      Inc(Dst);
    end;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.ApplyFloatFactors(Source: TPhotoBitmap);
// Apply FloatFactor scaling (Source supplies the factors)
var
  I: Integer;
  Src: PFloatColor;
  Dst: PFloatColor;
begin
  if (FFloatColors <> nil) and (Source.Size = FSize) then
  begin
    Src := Source.FFloatColors;
    Dst := FFloatColors;
    for I := 0 to FPixelCount - 1 do
    begin
      Dst^.fRed := Dst^.fRed * Src.fRed;
      if not FMonochrome then
      begin
        Dst^.fGreen := Dst^.fGreen * Src.fGreen;
        Dst^.fBlue := Dst^.fBlue * Src.fBlue;
      end
      else
      begin
        Dst^.fGreen := Dst^.fRed;
        Dst^.fBlue := Dst^.fRed;
      end;
      Inc(Src);
      Inc(Dst);
    end;
  end;
end;

// =============================================================================

procedure TPhotoBitmap.ShowFloatColors(Dest: TPhotoBitmap; Mult: Single = 1);
// Convert FFloatColors to RGB
var
  I: Integer;
  Col: Byte;
  R, G, B: Single;
  Dst: PRGB4;
  Src: PFloatColor;
begin
  MakeCompatible(Dest);
  if (FFloatColors <> nil) then
  begin
    Src := FFloatColors;
    Dst := Dest.Bits;
    for I := 0 to FPixelCount - 1 do
    begin
      R := Src^.fRed * Mult;
      Col := ClampByte(R);
      Dst^.rgbRed := Col;
      if not FMonochrome then
      begin
        G := Src^.fGreen * Mult;
        B := Src^.fBlue * Mult;
        Dst^.rgbBlue := ClampByte(B);
        Dst^.rgbGreen := ClampByte(G);
      end
      else
      begin
        Dst^.rgbBlue := Col;
        Dst^.rgbGreen := Col;
      end;
      Inc(Src);
      Inc(Dst);
    end;
  end;
end;

// =============================================================================

// Add SER frame
procedure TPhotoBitmap.AddSERFrame;
var
  I: Integer;
  Col: TFloatColor;
  Ptr: PWord;
  WR, WG, WB: Word;
  Max: Integer;
begin
  Max := Round(Power(2, FPixelDepth)) - 1;
  Ptr := FSERFile.AddFrame;
  for I := 0 to FPixelCount - 1 do
  begin
    Col := PFloatColor(Integer(FFloatColors) + (I * SizeOf(TFloatColor)))^;
    WR := Round(Col.fRed);
    if FMonochrome then
    begin
      Ptr^ := ClampWord(WR, Max);
      Inc(Ptr);
    end
    else
    begin
      WG := Round(Col.fGreen);
      WB := Round(Col.fBlue);
      Ptr^ := ClampWord(WG, Max);
      Inc(Ptr);
      Ptr^ := ClampWord(WB, Max);
      Inc(Ptr);
    end;
  end;
end;

// =============================================================================

// Clear SER frames
procedure TPhotoBitmap.ClearSERFrames;
begin
  FSERFile.DeleteFrames;
end;

// =============================================================================

// Get SER frame into FloatColors
function TPhotoBitmap.GetSERFrame(FrameNo: Integer; Dest: TPhotoBitmap)
  : Boolean;
var
  I: Integer;
  Col: TFloatColor;
  Src: PWord;
  Dst: PFloatColor;
begin
  Result := False;
  MakeCompatible(Dest);
  if (FSERFile.FrameCount > FrameNo) then
  begin
    Src := FSERFile.GetFrame(FrameNo);
    Dst := Dest.FloatColors;
    for I := 0 to FPixelCount - 1 do
    begin
      Col.fRed := Src^;
      Inc(Src);
      if not FMonochrome then
      begin
        Col.fGreen := Src^;
        Inc(Src);
        Col.fBlue := Src^;
        Inc(Src);
      end
      else
      begin
        Col.fGreen := Col.fRed;
        Col.fBlue := Col.fRed;
      end;
      Dst^ := Col;
      Inc(Dst);
    end;
    Result := True;
  end;
end;

// =============================================================================

// Save SER file
procedure TPhotoBitmap.SaveSERFile(FN: string);
begin
  FSERFile.SaveToFile(FN);
end;

// =============================================================================

// TPhoto

// =============================================================================

constructor TPhoto.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csOpaque];
  FBitmap := TPhotoBitmap.Create;
  FLayerCount := 1;
  FColor := clWhite;
  FTransColor := clTransBlack;
  SetLength(FLayers, 1);
  FLayers[0] := TPhotoBitmap.Create;
  FLayers[0].FColor := clWhite;
  FLayers[0].FTransparentColor := clTransBlack;
  FLayers[0].Height := 100;
  FLayers[0].Width := 100;
  FLayers[0].Parent := Self;
  Height := 100;
  Width := 100;
end;

// =============================================================================

destructor TPhoto.Destroy;
var
  I: Integer;
begin
  // Clean up
  FBitmap.Free;
  for I := 0 to FLayerCount - 1 do
    FreeAndNil(FLayers[I]);
  Finalize(FLayers);
  inherited Destroy;
end;

// =============================================================================

procedure TPhoto.WndProc(var Message: TMessage);
{ Intercept all Windows messages to the control }
begin
  case Message.Msg of
    WM_ERASEBKGND:
      if (NoErase) and not(csDesigning in ComponentState) then
      { Pretend we've done it }
      begin
        Message.Result := 1;
        Exit;
      end;
    WM_MOVE:
      NoErase := False;
    WM_SIZE:
      begin
        NoErase := False;
      end;
  end;
  inherited WndProc(Message);
end;

// =============================================================================

procedure TPhoto.Paint;
var
  J, X: Integer;
  W, H: Integer;
  TC: Cardinal;
  Src, Dst: PRGB4;
  Clr, Bkg: TRGB4;
  BO: TPoint;
begin
  if not FDeferPaint then
  begin
    if Assigned(FOnBeforePaint) then
      FOnBeforePaint(Self);
    NoErase := True;
    TC := RGBGR(FTransColor).Color;
    Bkg := RGBGR(FColor);
    W := 0;
    H := 0;
    for J := 0 to FLayerCount - 1 do
    begin
      if FLayers[J].Width > W then
        W := FLayers[J].Width;
      if FLayers[J].Height > H then
        H := FLayers[J].Height;
    end;
    FBitmap.Width := W;
    FBitmap.Height := H;
    Dst := PRGB4(FBitmap.FBits);
    for J := 0 to FLayerCount - 1 do
    begin
      if FLayers[J].Visible then
      begin
        Src := PRGB4(FLayers[J].FBits);
        if FLayers[J].Monochrome then
        begin
          for X := 0 to W * H - 1 do
          begin
            Clr := PRGB4(Integer(Src) + X * SizeOf(TRGB4))^;
            if (Clr.Color <> TC) then
            begin
              Clr.rgbGreen := Clr.rgbRed;
              Clr.rgbBlue := Clr.rgbRed;
              PRGB4(Integer(Dst) + X * SizeOf(TRGB4))^ := Clr;
            end;
          end;
        end
        else
        begin
          for X := 0 to W * H - 1 do
          begin
            Clr := PRGB4(Integer(Src) + X * SizeOf(TRGB4))^;
            if (Clr.Color <> TC) then
              PRGB4(Integer(Dst) + X * SizeOf(TRGB4))^ := Clr;
          end;
        end;
      end;
    end;
    if (FBitmap.Width = Width) and (FBitmap.Height = Height) then
    begin
      BitBlt(Canvas.Handle, 0, 0, Width, Height, FBitmap.Canvas.Handle, 0,
        0, SRCCOPY);
    end
    else
    begin
      SetStretchBltMode(Canvas.Handle, STRETCH_HALFTONE);
      SetBrushOrgEx(Canvas.Handle, 0, 0, @BO);
      StretchBlt(Canvas.Handle, 0, 0, Width, Height, FBitmap.Canvas.Handle, 0,
        0, FBitmap.Width, FBitmap.Height, SRCCOPY);
      SetBrushOrgEx(Canvas.Handle, BO.X, BO.Y, nil);
    end;
    if (csDesigning in ComponentState) then
      Canvas.DrawFocusRect(ClientRect);
    if Assigned(FOnAfterPaint) then
      FOnAfterPaint(Self);
  end;
end;

// =============================================================================

procedure TPhoto.BeginUpdate;
begin
  FDeferPaint := True;
end;

// =============================================================================

procedure TPhoto.EndUpdate;
begin
  FDeferPaint := False;
  Paint;
end;

// =============================================================================

procedure TPhoto.SetLayerCount(Value: Integer);
var
  I: Integer;
begin
  if (Value < 1) then
    Value := 1;
  for I := FLayerCount - 1 downto 0 do
    FreeAndNil(FLayers[I]);
  SetLength(FLayers, Value);
  for I := 0 to Value - 1 do
  begin
    FLayers[I] := TPhotoBitmap.Create;
    FLayers[I].Parent := Self;
    FLayers[I].FNoFloatColors := FNoFloatColors;
    if I = 0 then
      FLayers[I].FColor := FColor
    else
      FLayers[I].FColor := FTransColor;
    FLayers[I].FTransparentColor := FTransColor;
    FLayers[I].Height := Height;
    FLayers[I].Width := Width;
  end;
  FLayerCount := Value;
  Paint;
end;

// =============================================================================

procedure TPhoto.SetLayer(Idx: Integer; BMP: TPhotoBitmap);
begin
  FLayers[Idx].Assign(BMP);
end;

// =============================================================================

function TPhoto.GetLayer(Idx: Integer): TPhotoBitmap;
begin
  Result := FLayers[Idx];
end;

// =============================================================================

procedure TPhoto.SetBitmap(Value: TPhotoBitmap);
begin
  FBitmap.Assign(Value);
end;

// =============================================================================

procedure TPhoto.SetColor(Value: TColor);
begin
  if (FColor <> Value) then
  begin
    FColor := Value;
    Paint;
  end;
end;

// =============================================================================

procedure TPhoto.Show;
begin
  Visible := True;
end;

// =============================================================================

procedure TPhoto.Hide;
begin
  Visible := False;
end;

// =============================================================================

procedure TPhoto.WMGetDlgCode(var Msg: TWMGetDlgCode);
begin
  inherited;
  Msg.Result := DLGC_WANTARROWS;
end;

// =============================================================================

procedure TPhoto.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited;
end;

// =============================================================================

procedure TPhoto.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  inherited;
end;

// =============================================================================

procedure TPhoto.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited;
end;

// =============================================================================

function TPhoto.DoMouseWheel(Shift: TShiftState; WheelDelta: Integer;
  MousePos: TPoint): Boolean;
begin
  Result := inherited;
end;

// =============================================================================

function TPhoto.DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean;
begin
  Result := inherited;
end;

// =============================================================================

function TPhoto.DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean;
begin
  Result := inherited;
end;

// =============================================================================

// TSERFile

// =============================================================================

constructor TSERFile.Create;
begin
  inherited Create;
  FillChar(FHeader, SizeOf(TSERHeader), 0);
  FHeader.FileID := 'LUCAM-RECORDER';
  FFrames := TMemoryStream.Create;
end;

// =============================================================================

destructor TSERFile.Destroy;
begin
  FFrames.Free;
  inherited Destroy;
end;

// =============================================================================

procedure TSERFile.LoadFromFile(const FN: string);
begin
  FFrames.LoadFromFile(FN);
  CopyMemory(@FHeader, FFrames.Memory, SizeOf(TSERHeader));
end;

// =============================================================================

procedure TSERFile.SaveToFile(const FN: string);
begin
  if (FFrames.Size < SizeOf(TSERHeader)) then
    FFrames.SetSize(SizeOf(TSERHeader));
  CopyMemory(FFrames.Memory, @FHeader, SizeOf(TSERHeader));
  FFrames.SaveToFile(FN);
end;

// =============================================================================

procedure TSERFile.SetFileID(Value: AnsiString);
begin
  FillChar(FHeader.FileID, SizeOf(FHeader.FileID), #0);
  Move(PAnsiChar(Value)^, FHeader.FileID, Length(Value));
end;

function TSERFile.GetFileID: AnsiString;
begin
  Result := FHeader.FileID;
end;

// =============================================================================

procedure TSERFile.SetLuID(Value: Integer);
begin
  FHeader.LuID := Value;
end;

function TSERFile.GetLuID: Integer;
begin
  Result := FHeader.LuID;
end;

// =============================================================================

procedure TSERFile.SetColorID(Value: Integer);
begin
  FHeader.ColorID := Value;
  GetByteCount;
end;

function TSERFile.GetColorID: Integer;
begin
  Result := FHeader.ColorID;
end;

// =============================================================================

procedure TSERFile.SetLittleEndian(Value: Integer);
begin
  FHeader.LittleEndian := Value;
end;

function TSERFile.GetLittleEndian: Integer;
begin
  Result := FHeader.LittleEndian;
end;

// =============================================================================

procedure TSERFile.SetImageWidth(Value: Integer);
begin
  FHeader.ImageWidth := Value;
  GetByteCount;
end;

function TSERFile.GetImageWidth: Integer;
begin
  Result := FHeader.ImageWidth;
end;

// =============================================================================

procedure TSERFile.SetImageHeight(Value: Integer);
begin
  FHeader.ImageHeight := Value;
  GetByteCount;
end;

function TSERFile.GetImageHeight: Integer;
begin
  Result := FHeader.ImageHeight;
end;

// =============================================================================

procedure TSERFile.SetPixelDepth(Value: Integer);
begin
  FHeader.PixelDepth := Value;
  GetByteCount;
end;

function TSERFile.GetPixelDepth: Integer;
begin
  Result := FHeader.PixelDepth;
end;

// =============================================================================

procedure TSERFile.SetFrameCount(Value: Integer);
begin
  FHeader.FrameCount := Value;
  GetByteCount;
end;

function TSERFile.GetFrameCount: Integer;
begin
  Result := FHeader.FrameCount;
end;

// =============================================================================

procedure TSERFile.SetObserver(Value: AnsiString);
begin
  FillChar(FHeader.Observer, SizeOf(FHeader.Observer), #0);
  Move(PAnsiChar(Value)^, FHeader.Observer, Length(Value));
end;

function TSERFile.GetObserver: AnsiString;
begin
  Result := FHeader.Observer;
end;

// =============================================================================

procedure TSERFile.SetInstrument(Value: AnsiString);
begin
  FillChar(FHeader.Instrument, SizeOf(FHeader.Instrument), #0);
  Move(PAnsiChar(Value)^, FHeader.Instrument, Length(Value));
end;

function TSERFile.GetInstrument: AnsiString;
begin
  Result := FHeader.Instrument;
end;

// =============================================================================

procedure TSERFile.SetTelescope(Value: AnsiString);
begin
  FillChar(FHeader.Telescope, SizeOf(FHeader.Telescope), #0);
  Move(PAnsiChar(Value)^, FHeader.Telescope, Length(Value));
end;

function TSERFile.GetTelescope: AnsiString;
begin
  Result := FHeader.Telescope;
end;

// =============================================================================

procedure TSERFile.SetStartTime(Value: Int64);
begin
  FHeader.StartTime := Value;
end;

function TSERFile.GetStartTime: Int64;
begin
  Result := FHeader.StartTime;
end;

// =============================================================================

procedure TSERFile.SetStartTimeUTC(Value: Int64);
begin
  FHeader.StartTimeUTC := Value;
end;

function TSERFile.GetStartTimeUTC: Int64;
begin
  Result := FHeader.StartTimeUTC;
end;

// =============================================================================

function TSERFile.GetByteCount: Integer;
begin
  Result := FHeader.ImageWidth * FHeader.ImageHeight;
  if (FHeader.PixelDepth >= 8) then
    Result := Result * 2;
  if (FHeader.ColorID >= SER_RGB) then
    Result := Result * 3;
  FBytesPerFrame := Result;
  FBytesTotal := Result * FHeader.FrameCount;
end;

// =============================================================================

function TSERFile.AddFrame: Pointer;
begin
  FrameCount := FrameCount + 1;
  FFrames.SetSize(SizeOf(TSERHeader) + (FHeader.FrameCount * FBytesPerFrame));
  Result := Pointer(Integer(FFrames.Memory) + SizeOf(TSERHeader) +
    (FHeader.FrameCount - 1) * FBytesPerFrame);
end;

// =============================================================================

procedure TSERFile.DeleteFrames;
begin
  FrameCount := 0;
  FFrames.SetSize(SizeOf(TSERHeader));
end;

// =============================================================================

function TSERFile.GetFrame(FrameNo: Integer): Pointer;
begin
  Result := nil;
  if (FrameNo < FrameCount) then
  begin
    Result := Pointer(Integer(FFrames.Memory) + SizeOf(TSERHeader) + FrameNo *
      FBytesPerFrame);
  end;
end;

// =============================================================================

procedure Register;
begin
  RegisterComponents('Samples', [TPhoto]);
end;

// =============================================================================

end.
